
#pragma warning(push, 1)
#define WINVER         0x0400
#define _WIN32_WINNT   0x0400
#define _WIN32_WINDOWS 0x0400
#define _WIN32_IE      0x0400

#include <windows.h>
#include <gl/gl.h>
#pragma warning(pop)

#define GL_UNSIGNED_INT_8_8_8_8           0x8035
#define GL_BGRA                           0x80E1
#define GL_GENERATE_MIPMAP                0x8191
#define GL_UNSIGNED_INT_8_8_8_8_REV       0x8367

#define FAKE_FULLSCREEN 1
#define EXCLUSIVE_FULLSCREEN 2

typedef BOOL(WINAPI wglSwapInterval_t) (int interval);
wglSwapInterval_t *wglSwapInterval;
int swap_interval;

GLfloat wnd_width2;
int wnd_width;
int wnd_height;

int line_pos1;
int line_pos2;
int line_width;

GLfloat line1;
GLfloat line2;

int fullscreen;

WINDOWPLACEMENT wp;

void fake_fullscreen(HWND wnd)
{
	DWORD style = GetWindowLongA(wnd, GWL_STYLE);

	if (style & WS_OVERLAPPEDWINDOW && !fullscreen)
	{
		int screenw = GetSystemMetrics(SM_CXSCREEN);
		int screenh = GetSystemMetrics(SM_CYSCREEN);

		GetWindowPlacement(wnd, &wp);

		SetWindowLongA(wnd, GWL_STYLE, style & ~WS_OVERLAPPEDWINDOW);
		SetWindowPos(wnd, HWND_TOP, 0, 0, screenw, screenh,
			SWP_NOOWNERZORDER | SWP_FRAMECHANGED);

		fullscreen = FAKE_FULLSCREEN;
	}
	else if(fullscreen == FAKE_FULLSCREEN)
	{
		SetWindowLongA(wnd, GWL_STYLE, style | WS_OVERLAPPEDWINDOW);
		SetWindowPlacement(wnd, &wp);
		SetWindowPos(wnd, NULL, 0, 0, 0, 0,
			SWP_NOMOVE | SWP_NOSIZE | SWP_NOOWNERZORDER | SWP_FRAMECHANGED);

		fullscreen = 0;
	}
	
}

void exclusive_fullscreen(HWND wnd)
{
	DWORD style = GetWindowLongA(wnd, GWL_STYLE);

	if (style & WS_OVERLAPPEDWINDOW && !fullscreen)
	{
		DEVMODE dm;
		int     di = 0;
		int     found = 0;

		GetWindowPlacement(wnd, &wp);

		while(EnumDisplaySettings(NULL, di++, &dm))
		{
			if(dm.dmBitsPerPel >= 24 && dm.dmPelsWidth == 1024 &&
				dm.dmPelsHeight == 768 && dm.dmDisplayFrequency >= 60)
			{
				found = 1;
				break;
			}
		}

		if(!found)
			return;

		ChangeDisplaySettings(&dm, 0);

		SetWindowLongA(wnd, GWL_STYLE, style & ~WS_OVERLAPPEDWINDOW);
		SetWindowPos(wnd, HWND_TOP, 0, 0, dm.dmPelsWidth, dm.dmPelsHeight,
			SWP_NOOWNERZORDER | SWP_FRAMECHANGED);

		fullscreen = EXCLUSIVE_FULLSCREEN;
	}
	else if(fullscreen == EXCLUSIVE_FULLSCREEN)
	{
		ChangeDisplaySettings(NULL, 0);

		SetWindowLongA(wnd, GWL_STYLE, style | WS_OVERLAPPEDWINDOW);
		SetWindowPlacement(wnd, &wp);
		SetWindowPos(wnd, NULL, 0, 0, 0, 0,
			SWP_NOMOVE | SWP_NOSIZE | SWP_NOOWNERZORDER | SWP_FRAMECHANGED);

		fullscreen = 0;
	}
}

LRESULT CALLBACK wnd_proc(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	switch(msg)
	{
		case WM_CREATE:
			ShowWindow(wnd, SW_SHOW);
			return 0;
			
		case WM_SIZE:
			wnd_width   = LOWORD(lparam);
			wnd_height  = HIWORD(lparam);
			glViewport(0, 0, wnd_width, wnd_height);
			wnd_width2 = wnd_width * 0.5f;
			return 0;

		case WM_KEYUP:
			switch(wparam)
			{
				case VK_F1: fake_fullscreen(wnd); break;
				case VK_F2: exclusive_fullscreen(wnd); break;

				case VK_ADD:      wglSwapInterval(++swap_interval); break;
				case VK_SUBTRACT: wglSwapInterval(--swap_interval); break;
			}
			break;

		case WM_CLOSE:
			DestroyWindow(wnd);
			return 0;

		case WM_DESTROY:
			PostQuitMessage(0);
			return 0;
	}

	return DefWindowProcA(wnd, msg, wparam, lparam);
}

int WINAPI WinMain(HINSTANCE inst, HINSTANCE prev, char *cmdline, int cmdshow)
{
	PIXELFORMATDESCRIPTOR pfd = { sizeof(PIXELFORMATDESCRIPTOR) };

	HWND      wnd;
	WNDCLASSA wc;
	ATOM      rc;
	MSG       msg;
	int       running;

	int       pf;
	BOOL      spf;
	HDC       dev_ctx;
	HGLRC     ren_ctx;

	wp.length = sizeof(WINDOWPLACEMENT);

	wc.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wc.lpfnWndProc = wnd_proc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = inst;
	wc.hIcon = LoadIcon(NULL, IDI_WINLOGO);
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = NULL;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = "TEARING_TEST";

	rc = RegisterClassA(&wc);
	if (!rc) ExitProcess(1);

	wnd = CreateWindowA("TEARING_TEST", "Tearing Test", WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, CW_USEDEFAULT,
		640, 480, NULL, NULL, inst, NULL);
	if (!wnd) ExitProcess(2);

	dev_ctx = GetDC(wnd);
	if (!dev_ctx) ExitProcess(3);

	pfd.nVersion   = 1;
	pfd.dwFlags    = PFD_DOUBLEBUFFER | PFD_SUPPORT_OPENGL | PFD_DRAW_TO_WINDOW | PFD_DEPTH_DONTCARE;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.cColorBits = 24;
	pfd.cAlphaBits = 8;
	pfd.iLayerType = PFD_MAIN_PLANE;

	pf = ChoosePixelFormat(dev_ctx, &pfd);
	if (!pf) ExitProcess(4);

	spf = SetPixelFormat(dev_ctx, pf, &pfd);
	if (!spf) ExitProcess(5);

	ren_ctx = wglCreateContext(dev_ctx);
	if (ren_ctx == 0) ExitProcess(6);

	wglMakeCurrent(dev_ctx, ren_ctx);

	wglSwapInterval = (wglSwapInterval_t*)wglGetProcAddress("wglSwapIntervalEXT");
	if (!wglSwapInterval) ExitProcess(7);

	line_width = 4;
	glLineWidth((GLfloat)line_width);
	glClearColor(0.20f, 0.20f, 0.20f, 1.0f);
	//wglSwapInterval(1);

	line_pos1 = 0;
	line_pos2 = line_width * 4;

	running = TRUE;
	while (running)
	{
		while (PeekMessageA(&msg, NULL, 0, 0, PM_REMOVE))
		{
			if (msg.message == WM_QUIT)
				running = FALSE;

			TranslateMessage(&msg);
			DispatchMessageA(&msg);
		}

		glClear(GL_COLOR_BUFFER_BIT);

		line_pos1 += line_width;
		line_pos2 += line_width;

		line1 = (line_pos1 % wnd_width) / wnd_width2 - 1.0f;
		line2 = (line_pos2 % wnd_width) / wnd_width2 - 1.0f;

		glBegin(GL_LINES);
			glColor3f(1.0f, 0.0f, 0.0f);
			glVertex2f(line1,  1.0f);
			glVertex2f(line1, -1.0f);
			glVertex2f(line2,  1.0f);
			glVertex2f(line2, -1.0f);
		glEnd();

		SwapBuffers(dev_ctx);
	}

	return 0;
}